/* Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_TRI

//`define OUTPUT_MUX_TRI

// Module Name: rtavr_ior_spi
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                          DO:  NORMAL    TRISTATE(FAKE)   
// Number of Slice Flip Flops:     16         16
// Number of 4 input LUTs:         22         23
// Number of occupied Slices:      19         20
// Total Number of 4 input LUTs:   22         23 
// Number of bonded IOBs :         33         33
//   IOB Flip Flops :              16         16
// Number of BUFGMUXs :             1          1
// Number of RAMB16BWEs :          -           -

module rtavr_ior_spi(
		input CLK,
		input RESET,

		input [7:0] DI,
		output [7:0] DO
		, input WR,RD
		, input [5:0] ADDR
		// external I/O
		, inout MISO
		, inout MOSI
		, inout SC
		, inout SS
		// interrupt
		, output INT_SPI
		, input RST_INT_SPI
		);
  parameter BASE_ADDR  = 6'h2e;    // SPI : SPDR 0x2e, SPSR, SPCR

  reg  [7:0] r_spdr;
  reg  [7:0] r_spcr;
  reg  r_spif; // spsr bit7
  reg  r_wcol; // spsr bit6
  reg  r_spi2x; // spsr bit0

  wire SPIE = r_spcr[7];
  wire SPE = r_spcr[6];
  wire DORD = r_spcr[5];
  wire MSTR = r_spcr[4];
  wire CPOL = r_spcr[3];
  wire CPHA = r_spcr[2];
  wire [1:0] SPR = r_spcr[1:0];

  wire OE_spdr = RD & (ADDR == BASE_ADDR);
  wire WE_spdr = WR & (ADDR == BASE_ADDR);
  wire OE_spsr = RD & (ADDR == (BASE_ADDR + 1)); 
  wire WE_spsr = WR & (ADDR == (BASE_ADDR + 1));
  wire OE_spcr = RD & (ADDR == (BASE_ADDR + 2)); 
  wire WE_spcr = WR & (ADDR == (BASE_ADDR + 2));

  wire [7:0] i_DO = OE_spdr ? r_spdr[7:0] 
		: OE_ubrrh ? { r_spif, r_wcol, 5'b0, r_spi2x }
		: OE_spcr ? r_spcr: 0;
`ifdef OUTPUT_MUX_TRI
  bufif1(DO, i_DO, (OE_ubrrl | OE_ubrrh | OE_ucsrb| OE_ucsra | OE_udr));
`else
  assign DO = i_DO;
`endif
  always @(negedge CLK)
  begin
    if (RESET == 0)
      begin
      end
    else
      begin
      end
  end
  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	r_spdr <= 0;
	r_spcr <= 0;
	r_spif <= 0;
	r_wcol <= 0;
	r_spi2x <= 0;
      end
    else
      begin
	if(WE_spdr) r_spdr <= DI;
	if(WE_spcr) r_spcr <= DI;
	if(WE_spsr) r_spif <= DI[7];
	if(WE_spsr) r_wcol <= DI[6];
	if(WE_spsr) r_spi2x <= DI[0];
      end
  end

endmodule
