/* Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_TRI

//`define OUTPUT_MUX_TRI

// Module Name: rtavr_ior_usart
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                          DO:  NORMAL    TRISTATE(FAKE)   
// Number of Slice Flip Flops:     16         16
// Number of 4 input LUTs:         22         23
// Number of occupied Slices:      19         20
// Total Number of 4 input LUTs:   22         23 
// Number of bonded IOBs :         33         33
//   IOB Flip Flops :              16         16
// Number of BUFGMUXs :             1          1
// Number of RAMB16BWEs :          -           -

module rtavr_ior_usart(
		input CLK,
		input RESET,

		input [7:0] DI,
		output [7:0] DO
		, input WR,RD
		, input [5:0] ADDR
		// external I/O
		, output TXD
		, input RXD
		, inout XCK
		// interrupt
		, output INT_TX
		, input RST_INT_TX
		, output INT_UDRE
		, input RST_INT_UDRE
		, output INT_RX
		, input RST_INT_RX
		);
  parameter BASE_ADDR = 6'h0e;   // USART : (0x0e - 0x14 ADC space)
					UBRRL, UBRRH, UCSRC, UCSRB, UCSRA, UDR

  reg  [7:0] r_udr_tx0;
  reg  [7:0] r_udr_tx1;
  reg  [7:0] r_udr_rx0;
  reg  [7:0] r_udr_rx1;
  reg  [7:0] r_ucsra;
  reg  [7:0] r_ucsrb;
  reg  [6:0] r_ucsrc;
  reg  [11:0] r_ubrr;

  wire RXC = r_ucsra[7];
  wire TXC = r_ucsra[6];
  wire UDRE = r_ucsra[5];
  wire FE = r_ucsra[4];
  wire DOR = r_ucsra[3];
  wire PE = r_ucsra[2];
  wire U2X = r_ucsra[1];
  wire MPCM = r_ucsra[0];

  wire RXCIE  =  r_ucsrb[7];
  wire TXCIE  =  r_ucsrb[6];
  wire UDRIE  =  r_ucsrb[5];
  wire RXEN   =  r_ucsrb[4];
  wire TXEN   =  r_ucsrb[3];
//  wire RXB8 =  r_ucsrb[1];
//  wire TXB8 = r_ucsrb[0]; 

  wire UMSEL      = r_ucsrc[6];
  wire [1:0] UPM  = r_ucsrc[5:4];
  wire USBS       = r_ucsrc[3];
  wire [2:0] UCSZ = { r_ucsrb[2], r_ucsrc[2:1]};
  wire UCPOL      = r_ucsrc[0];

  wire OE_ubrrl = RD & (ADDR == BASE_ADDR);
  wire WE_ubrrl = WR & (ADDR == BASE_ADDR);
  wire OE_ubrrl = RD & (ADDR == (BASE_ADDR + 1)); 
  wire WE_ubrrh = WR & (ADDR == (BASE_ADDR + 1));
  wire OE_ucsrc = RD & (ADDR == (BASE_ADDR + 2)); 
  wire WE_ucsrc = WR & (ADDR == (BASE_ADDR + 2));
  wire OE_ucsrb = RD & (ADDR == (BASE_ADDR + 3)); 
  wire WE_ucsrb = WR & (ADDR == (BASE_ADDR + 3));
  wire OE_ucsra = RD & (ADDR == (BASE_ADDR + 4)); 
  wire WE_ucsra = WR & (ADDR == (BASE_ADDR + 4));
  wire OE_udr   = RD & (ADDR == (BASE_ADDR + 5)); 
  wire WE_udr   = WR & (ADDR == (BASE_ADDR + 5));

`ifdef OUTPUT_MUX_TRI
  wire [7:0] i_DO = OE_ubrrl ? r_ubrr[7:0] 
		: OE_ubrrh ? { 4'h0 , r_ubrr[11:8] }
		: OE_ucsrb ? r_ucsrb   : OE_ucsra ? r_ucsra
		: OE_ucsrc ? { 1'b0 , r_ucsrc[6:0] }
		: OE_udr ? r_udr_rx0 : 0;
  bufif1(DO, i_DO, (OE_ubrrl | OE_ubrrh | OE_ucsrb| OE_ucsra | OE_udr));
`else
  wire [7:0] i_DO = OE_ubrrl ? r_ubrr[7:0] 
		: OE_ubrrh ? { 4'h0 , r_ubrr[11:8] }
		: OE_ucsrb ? r_ucsrb  : OE_ucsra ? r_ucsra
		: OE_ucsrc ? { 1'b0 , r_ucsrc[6:0] }
		: OE_udr ? r_udr_rx0 : 0;
  assign DO = i_DO;
`endif
  always @(negedge CLK)
  begin
    if (RESET == 0)
      begin
      end
    else
      begin
      end
  end
  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	r_usbrr <= 0;
	r_ucsra <= 0;
	r_ucsrb <= 0;
	r_ucsrc <= 0;
	r_udr_rx0 <= 0;
	r_udr_rx1 <= 0;
	r_udr_tx0 <= 0;
	r_udr_tx1 <= 0;
      end
    else
      begin
	if(WE_ubrrl) r_ubrr[7:0] <= DI;
	if(WE_ubrrh) r_ubrr[11:8] <= DI[3:0];
	if(WE_ucsra) r_ucsra <= DI;
	if(WE_ucsrb) r_ucsrb <= DI;
	if(WE_ucsrc) r_ucsrb <= DI[6:0];
	if(WE_udr)   r_udr_tx1 <= DI;
      end
  end

endmodule
