/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_TRI, IOR_HAVE_PORTA , IOR_HAVE_PORTB
//                 , IOR_HAVE_PORTC, IOR_HAVE_INT0, IOR_HAVE_SREG_SP

//`define OUTPUT_MUX_TRI
//`define IOR_HAVE_PORTA
//`define IOR_HAVE_PORTB
//`define IOR_HAVE_PORTC
//`define IOR_HAVE_INT0
//`define IOR_HAVE_TIM0

//`define IOR_HAVE_SREG_SP

// Module Name: rtavr_ior
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//                              NO_PORT  +PORTA +PORTB +PORTC +SREG/SP
// Number of Slice Flip Flops:     61      77     93     112     136
// Number of 4 input LUTs:         97     125    143     176     223
// Number of occupied Slices:      81     104    122     152     175
// Total Number of 4 input LUTs:  106     134    152     185     232
// Number of bonded IOBs :         34      42     50      58     106
//   IOB Flip Flops :               -      16     32      47      71
// Number of BUFGMUXs :             1       1      1       1       1
// Number of RAMB16BWEs :           -       -      -       -       -
// CLK to Setup                 5.921                          5.962
// PORTC -- PORTC + INT0

module rtavr_ior (
		input CLK,
		input RESET,

		input WR,RD, 
		input [7:0] DI,
		input [5:0] ADDR,
		output [7:0] DO

`ifdef IOR_HAVE_SREG_SP
		, input  [7:0] FLAGS_IN
		, input  [15:0] SP_IN
		, output [7:0] FLAGS_OUT
		, output [15:0] SP_OUT
`endif
		// external I/O
`ifdef IOR_HAVE_PORTA
		, inout [7:0] PORTA
`endif
`ifdef IOR_HAVE_PORTB
		, inout [7:0] PORTB
`endif
`ifdef IOR_HAVE_PORTC
		, inout [7:0] PORTC
`endif

		// interrupt handling
`ifdef IOR_HAVE_INT0
		, input  RST_INT0
		, output INT0
`endif

`ifdef IOR_HAVE_TIM0
		, input  RST_TIM0_OVF
		, output TIM0_OVF
		, input  RST_TIM0_COMPA	
		, output TIM0_COMPA	
		, input  RST_TIM0_COMPB
		, output TIM0_COMPB
`endif
		);
  
  parameter PORTA_ADDR = 6'h00;   // PINA 0x00 , DDRA, PORTA
  parameter PORTB_ADDR = 6'h04;   // PINB 0x04 , DDRB, PORTB
  parameter PORTC_ADDR = 6'h1b;   // PINC 0x1b , DDRC, PORTC
  parameter TIM0_ADDR  = 6'h15;   // OCR0B 0x15 - TCCR0A 0x19
  parameter SREG_ADDR  = 6'h3f;   // SREG 0x3f , SPH,SPL,CCP,RSRFR,MCUCR 0x3a
  parameter MCUCR_ADDR = 6'h3a;   // MCUCR 0x3a
  parameter GIFR_ADDR  = 6'h0b;   // GIFR 0x0b , GIMSK 0x0c
  parameter TIFR_ADDR  = 6'h25;   // TIFR 0x25 , TIMSK 0x26  
  parameter INT0_BIT = 18;      // PC2
  // parameter INT0_BIT = 0;      // PA0 : for test
  parameter USART_ADDR = 6'h0e;   // USART : (0x0e - 0x14 ADC space)
				  //	UBRRL, UBRRH, UCSRC, UCSRB, UCSRA, UDR
  parameter SPI_ADDR  = 6'h2e;    // SPI : SPDR 0x2e, SPSR, SPCR
  parameter TWI_ADDR  = 6'h28;    // TWI : TWSD 0x28, TWSAM,TWSA, TWSSRA
				  //		,TWSCRB, TWSCRA

`ifdef IOR_HAVE_SREG_SP
  reg  [7:0] r_sreg;
  reg  [7:0] r_sph;
  reg  [7:0] r_spl;

  assign FLAGS_OUT = r_sreg;
  assign SP_OUT  = { r_sph , r_spl };

  wire OE_sreg = RD & (ADDR == SREG_ADDR); 
  wire WE_sreg = WR & (ADDR == SREG_ADDR);
  wire OE_sph = RD & (ADDR == (SREG_ADDR -1)); 
  wire WE_sph = WR & (ADDR == (SREG_ADDR -1));
  wire OE_spl = RD & (ADDR == (SREG_ADDR -2)); 
  wire WE_spl = WR & (ADDR == (SREG_ADDR -2));
`else
  wire  [7:0] r_sreg = 0;
  wire  [7:0] r_sph = 0;
  wire  [7:0] r_spl = 0;

  wire OE_sreg = 0;
  wire OE_sph = 0;
  wire OE_spl = 0;
`endif

`ifdef IOR_HAVE_TIM0
  wire OE_tifr = RD & (ADDR == TIFR_ADDR); 
  wire WE_tifr = WR & (ADDR == TIFR_ADDR);
  wire OE_timsk = RD & (ADDR == (TIFR_ADDR +1)); 
  wire WE_timsk = WR & (ADDR == (TIFR_ADDR +1));

  reg t0_ovf_prev;  
  reg t0_compa_prev;
  reg t0_compb_prev;  
  reg t0_ovf, t0_ovf_ie; // bit0
  reg t0_compa, t0_compa_ie; // bit1
  reg t0_compb, t0_compb_ie; // bit2

  assign TIM0_OVF   = t0_ovf & t0_ovf_ie;       // int 10
  assign TIM0_COMPA = t0_compa & t0_compa_ie ;  // int 11
  assign TIM0_COMPB = t0_compb & t0_compb_ie ;  // int 12
`else
  wire OE_tifr = 0; 
  wire WE_tifr = 0;
  wire OE_timsk = 0; 
  wire WE_timsk = 0;
`endif // IOR_HAVE_TIM0


  //reg intf0, int0; // bit0
`ifdef IOR_HAVE_INT0
  wire OE_gifr = RD & (ADDR == GIFR_ADDR); 
  wire WE_gifr = WR & (ADDR == GIFR_ADDR);
  wire OE_gimsk = RD & (ADDR == (GIFR_ADDR +1)); 
  wire WE_gimsk = WR & (ADDR == (GIFR_ADDR +1));
  wire OE_mcucr = RD & (ADDR == MCUCR_ADDR); 
  wire WE_mcucr = WR & (ADDR == MCUCR_ADDR);

  wire [4:0] gi_int0_bit = INT0_BIT;
  reg gi_int0_in; 
  reg gi_int0_prev;  
  reg gi_int0, gi_int0_ie; // bit0
  reg [1:0] mcucr_isc;
  
  assign INT0 = gi_int0;
`else
  wire OE_gifr = 0; 
  wire WE_gifr = 0;
  wire OE_gimsk = 0; 
  wire WE_gimsk = 0;
  wire OE_mcucr = 0; 
  wire WE_mcucr = 0;
`endif


`ifdef OUTPUT_MUX_TRI
    rtavr_ior_port #(.BASE_ADDR(PORTA_ADDR)) i_port_a(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(DO) , .PORT(PORTA)  
		, .WR(WR), .RD(RD), .ADDR(ADDR) );

    rtavr_ior_port #(.BASE_ADDR(PORTB_ADDR)) i_port_b(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(DO) , .PORT(PORTB)  
		, .WR(WR), .RD(RD), .ADDR(ADDR) );

    rtavr_ior_port #(.BASE_ADDR(PORTC_ADDR)) i_port_c(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(DO) , .PORT(PORTC) 
		, .WR(WR), .RD(RD), .ADDR(ADDR) );


`ifdef IOR_HAVE_TIM0
    rtavr_ior_timer0 #(.BASE_ADDR(TIM0_ADDR)) i_timer0(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(DO)
		, .WR(WR), .RD(RD), .ADDR(ADDR)
		, .TIM0_OVF(TIM0_OVF_IN)
		, .TIM0_COMPB(TIM0_COMPB_IN)
		, .TIM0_COMPA(TIM0_COMPA_IN) );
`endif

  wire [7:0] i_DO = OE_sreg ? r_sreg : OE_sph ? r_sph : OE_spl ? r_spl 
                  : OE_tifr ? {5'b0 , t0_compb, t0_compa ,t0_ovf}
		  : OE_timsk ? {5'b0 , t0_compb_ie, t0_compa_ie ,t0_ovf_ie}
`ifdef IOR_HAVE_INT0
                  : OE_gifr ? {7'b0 , gi_int0}
		  : OE_gimsk ? {7'b0 , gi_int0_ie}
		  : OE_mcucr ? {mcucr_isc , 6'b0 }
`endif
		  : 0;

  bufif1(DO, i_DO, (OE_sreg | OE_sph | OE_spl | OE_tifr | OE_timsk ));

`else // not OUTPUT_MUX_TRI

  wire [7:0] pa_DO;
  wire [7:0] pb_DO;
  wire [7:0] pc_DO;
  wire [7:0] t0_DO;
  
`ifdef IOR_HAVE_PORTA
    wire OE_pa = RD & ( (ADDR == PORTA_ADDR) | (ADDR == PORTA_ADDR+1)
		     | (ADDR == PORTA_ADDR+2) );
    rtavr_ior_port #(.BASE_ADDR(PORTA_ADDR)) i_port_a(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(pa_DO) , .PORT(PORTA)  
		, .WR(WR), .RD(RD), .ADDR(ADDR) );
`else
    assign pa_DO = 8'h00;
    wire OE_pa = 0;
`endif
`ifdef IOR_HAVE_PORTB
    wire OE_pb = RD & ( (ADDR == PORTB_ADDR) | (ADDR == PORTC_ADDR+1)
		     | (ADDR == PORTB_ADDR+2) );
    rtavr_ior_port #(.BASE_ADDR(PORTB_ADDR)) i_port_b(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(pb_DO) , .PORT(PORTB)  
		, .WR(WR), .RD(RD), .ADDR(ADDR) );
`else
    assign pb_DO = 8'h00;
    wire OE_pb = 0;
`endif

`ifdef IOR_HAVE_PORTC
    wire OE_pc = RD & ( (ADDR == PORTC_ADDR) | (ADDR == PORTC_ADDR+1)
		     | (ADDR == PORTC_ADDR+2) );
    rtavr_ior_port #(.BASE_ADDR(PORTC_ADDR)) i_port_c(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(pc_DO) , .PORT(PORTC) 
		, .WR(WR), .RD(RD), .ADDR(ADDR) );
`else
    assign pc_DO = 8'h00;
    wire OE_pc = 0;
`endif

`ifdef IOR_HAVE_TIM0
   rtavr_ior_timer0 #(.BASE_ADDR(TIM0_ADDR)) i_timer0(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(t0_DO)
		, .WR(WR), .RD(RD), .ADDR(ADDR)
		, .TIM0_OVF(TIM0_OVF_IN)
		, .TIM0_COMPB(TIM0_COMPB_IN)
		, .TIM0_COMPA(TIM0_COMPA_IN) );
   wire OE_t0 = RD & ( (ADDR == TIM0_ADDR) | (ADDR == TIM0_ADDR+1)
                     | (ADDR == TIM0_ADDR+2) | (ADDR == TIM0_ADDR+3)
		     | (ADDR == TIM0_ADDR+4) );

`else
    assign t0_DO = 8'h00;
    wire OE_t0 = 0;
`endif

   wire [7:0] i_DO = OE_sreg ? r_sreg : OE_sph ? r_sph : OE_spl ? r_spl
`ifdef IOR_HAVE_TIM0
                  : OE_tifr ? {5'b0 , t0_compb, t0_compa ,t0_ovf}
		  : OE_timsk ? {5'b0 , t0_compb_ie, t0_compa_ie ,t0_ovf_ie}
`endif
`ifdef IOR_HAVE_INT0
                  : OE_gifr ? {7'b0 , gi_int0}
		  : OE_gimsk ? {7'b0 , gi_int0_ie}
		  : OE_mcucr ? {mcucr_isc , 6'b0 }
`endif

		  : 0;
  `ifdef OUTPUT_MUX_OR
   assign DO =  (OE_sreg | OE_sph | OE_spl | OE_tifr | OE_timsk ) ? i_DO
              : pa_DO | pb_DO | pc_DO | t0_DO;   
  `else
   assign DO =  (OE_sreg | OE_sph | OE_spl | OE_tifr | OE_timsk ) ? i_DO
              : OE_pa ? pa_DO : OE_pb ? pb_DO : OE_pc ? pc_DO
	      : OE_t0 ? t0_DO : 8'h00;   
  `endif

`endif // not OUTPUT_MUX_TRI

  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
`ifdef IOR_HAVE_SREG_SP
	r_sreg <= 0;
	r_sph <= 0;
	r_spl <= 0;
`endif
`ifdef IOR_HAVE_TIM0
	t0_compb_ie <= 0;
	t0_compa_ie <= 0;
	t0_ovf_ie <= 0;

	t0_ovf <= 0;
	t0_compa <= 0;

	t0_compb <= 0;

	t0_ovf_prev   <=  0;
	t0_compa_prev <=  0;
	t0_compb_prev <=  0;
`endif

`ifdef IOR_HAVE_INT0
	mcucr_isc <= 0;

	gi_int0_ie <= 0;
	gi_int0 <= 0;
	gi_int0_in   <= 0;
	gi_int0_prev   <=  0;
`endif
      end
    else
      begin
`ifdef IOR_HAVE_SREG_SP
	if(WE_sreg) r_sreg <= DI;
	else r_sreg <= FLAGS_IN;

	if(WE_sph) r_sph <= DI;
	else r_sph <= SP_IN[15:8];

	if(WE_spl) r_spl <= DI;
	else r_spl <= SP_IN[7:0];
`endif
`ifdef IOR_HAVE_TIM0
	if(WE_timsk) {t0_compb_ie, t0_compa_ie ,t0_ovf_ie} <= DI[2:0];

	if(RST_TIM0_OVF) t0_ovf <= 0;
	else if((WE_tifr & DI[0]) 
             | (t0_ovf_prev ^ TIM0_OVF_IN) & ~t0_ovf_prev ) t0_ovf <= 1;

	if(RST_TIM0_COMPA) t0_compa <= 0;
	else if((WE_tifr & DI[1])
             | (t0_compa_prev ^ TIM0_COMPA_IN) & ~t0_compa_prev) t0_compa <= 1;

	if(RST_TIM0_COMPB) t0_compb <= 0;
	else if((WE_tifr & DI[2])
             | (t0_compb_prev ^ TIM0_COMPB_IN) & ~t0_compb_prev) t0_compb <= 1;

	t0_ovf_prev   <=  TIM0_OVF_IN;
	t0_compa_prev <=  TIM0_COMPA_IN;
	t0_compb_prev <=  TIM0_COMPB_IN; 
`endif //  IOR_HAVE_TIM0

`ifdef IOR_HAVE_INT0
	if(WE_mcucr) {mcucr_isc} <= DI[7:6];

	if(WE_gimsk) {gi_int0_ie} <= DI[0:0];
	if(RST_INT0) gi_int0 <= 0;
	else if((WE_gifr & DI[0])
              | (gi_int0_prev ^ gi_int0_in) 
                 & ((mcucr_isc == 1) 
                   | (mcucr_isc == 2) & ( gi_int0_prev) 
                   | (mcucr_isc == 3) & ~( gi_int0_prev) )  
	       | ((mcucr_isc == 0) & ~gi_int0_in) ) gi_int0 <= 1;

	gi_int0_in   <= 
  `ifdef IOR_HAVE_PORTA
	( gi_int0_bit[4:3] == 0) ? PORTA[gi_int0_bit[2:0]] :
  `endif
  `ifdef IOR_HAVE_PORTB
	( gi_int0_bit[4:3] == 1) ? PORTB[gi_int0_bit[2:0]] :
  `endif
  `ifdef IOR_HAVE_PORTC
	( gi_int0_bit[4:3] == 2) ? PORTC[gi_int0_bit[2:0]] :
  `endif
	1'b0;

	gi_int0_prev   <=  gi_int0_in;
`endif // IOR_HAVE_INT0
      end
  end

endmodule
